import torch.nn as nn
import ltr.models.backbone as backbones
import ltr.models.bbreg as bbmodels
import ltr.models.locator as locmodels
from ltr import model_constructor

class SBDTNet(nn.Module):
    """ SBDTNet network module"""
    def __init__(self, feature_extractor, bb_regressor, bb_regressor_layer, location_predictor, location_predictor_layer, extractor_grad):
        """
        args:
            feature_extractor - backbone feature extractor
            bb_regressor - IoU prediction module
            bb_regressor_layer - List containing the name of the layers from feature_extractor, which are input to bb_regressor
            extractor_grad - Bool indicating whether backbone feature extractor requires gradients
            location_predictor - location prediction module
            location_predictor_layer - List containing the name of the layers from feature_extractor, which are input to location_predictor
        """
        super(SBDTNet, self).__init__()

        self.feature_extractor = feature_extractor
        self.bb_regressor = bb_regressor
        self.bb_regressor_layer = bb_regressor_layer
        self.location_predictor = location_predictor
        self.location_predictor_layer = location_predictor_layer

        if not extractor_grad:
            for p in self.feature_extractor.parameters():
                p.requires_grad_(False)

    def forward(self, train_imgs, test_imgs, train_bb, test_scale_proposals, train_locator_proposals, train_locator_labels, test_locator_proposals):
        pass

    def extract_backbone_features(self, im, layers=None):
        if layers is None:
            layers = list(set(self.bb_regressor_layer + self.location_predictor_layer))
        return self.feature_extractor(im, layers)

    def extract_features(self, im, layers):
        return self.feature_extractor(im, layers)


@model_constructor
def SBDT_resnet18(iou_input_dim=(256,256), iou_inter_dim=(256,256), backbone_pretrained=True):
    # backbone
    backbone_net = backbones.resnet18(pretrained=backbone_pretrained)

    # Bounding box regressor
    iou_predictor = bbmodels.AtomIoUNet(pred_input_dim=iou_input_dim, pred_inter_dim=iou_inter_dim)

    # locator
    location_predictor = locmodels.OnlineRRNet()

    net = SBDTNet(feature_extractor=backbone_net, bb_regressor=iou_predictor, bb_regressor_layer=['layer2', 'layer3'], \
                  location_predictor=location_predictor, location_predictor_layer=['layer2', 'layer3'], extractor_grad=False)

    return net
