import os
import torch
import torchvision
import numpy as np

from ltr import load_network
from pytracking import TensorList
from pytracking.features.featurebase import FeatureBase, MultiFeatureBase

normalize = torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

class DCFSTNet18:
	"""DCFSTNet18 with ResNet18 backbone.
	args:
		net_path: Relative or absolute net path.
		use_gpu: Use GPU or CPU.
	"""
	def __init__(self, net_path, use_gpu=True, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.use_gpu = use_gpu
		self.net_path = net_path

	def initialize(self):
		if os.path.isabs(self.net_path):
			net_path_full = self.net_path
		else:
			net_path_full = os.path.join(os.path.dirname(__file__), '..', 'networks', self.net_path)

		self.net, _ = load_network(net_path_full, backbone_pretrained=False)

		self.net.cuda()
		self.net.eval()

		self.mean = torch.Tensor([0.485, 0.456, 0.406]).view(1,-1,1,1).cuda()
		self.std = torch.Tensor([0.229, 0.224, 0.225]).view(1,-1,1,1).cuda()

		self.layer_stride = {'conv1': 2, 'layer1': 4, 'layer2': 8, 'layer3': 16, 'layer4': 32, 'classification': 16, 'fc': None}
		self.layer_dim = {'conv1': 64, 'layer1': 64, 'layer2': 128, 'layer3': 256, 'layer4': 512, 'classification': 256,'fc': None}

		self.iou_predictor = self.net.bb_regressor
		self.location_predictor = self.net.location_predictor

		self.iounet_feature_layers = self.net.bb_regressor_layer
		self.locator_feature_layers = self.net.location_predictor_layer
		self.backbone_feature_layers = sorted(list(set(self.iounet_feature_layers + self.locator_feature_layers)))

		## all parameters do not require grad
		for p in self.net.parameters():
			p.requires_grad = False

	def free_memory(self):
		if hasattr(self, 'net'):
			del self.net
		if hasattr(self, 'iou_predictor'):
			del self.iou_predictor
		if hasattr(self, 'iounet_backbone_features'):
			del self.iounet_backbone_features
		if hasattr(self, 'iounet_features'):
			del self.iounet_features
		if hasattr(self, 'location_predictor'):
			del self.location_predictor
		if hasattr(self, 'locator_backbone_features'):
			del self.locator_backbone_features
		if hasattr(self, 'locator_features'):
			del self.locator_features

	def extract(self, im: torch.Tensor, locator_proposals: torch.Tensor, only_locator=False):
		if im.is_cuda == False:
			im = im.cuda()
		if im.dim() == 3:
			im = im.unsqueeze(0)

		im = self.preprocessing(im)

		with torch.no_grad():
			backbone_features = self.net.extract_backbone_features(im, self.backbone_feature_layers)

		# locator features
		self.locator_backbone_features = TensorList([backbone_features[layer].clone() for layer in self.locator_feature_layers])
		with torch.no_grad():
			self.locator_features = self.location_predictor.get_locator_feat(self.locator_backbone_features, locator_proposals).squeeze()

		# iou features
		if not only_locator:
			self.iounet_backbone_features = TensorList([backbone_features[layer].clone() for layer in self.iounet_feature_layers])
			with torch.no_grad():
				self.iounet_features = TensorList(self.iou_predictor.get_iou_feat(self.iounet_backbone_features))

	def preprocessing(self, im):
		im = im / 255
		im = im - self.mean
		im = im / self.std
		return im